from abaqus import *
from abaqusConstants import *
import math, sys, copy
from exceptions import *
from customKernel import RepositorySupport, RegisteredList, CommandRegister, RegisteredDictionary
from abq_Pme.pmeUtils import getDefaultPartSettings
from abq_Pme.pmeRegions import PmeRegions
import mesh
import regionToolset

from abq_Pme.pmeConstants import *

class PmeModel(PmeRegions, RepositorySupport, CommandRegister):
    def __init__(self, name):
        PmeRegions.__init__(self)
        RepositorySupport.__init__(self)
        CommandRegister.__init__(self)
        self.name = name
        self.columnWidths = ()

        # Local variables
        self.table = self.initializeTable()
        self.activationMap = self.updateActivationMap() # {partName:isActivated}
        self.prevSettings = self.getSettings(self.table) # Just a dictionary of self.table
        self.shellThicknessesToMaterialMap = {} # {materialName:{thickness:thickNumber}}
        self.shellThicknessesFromSolids = {} # Stored thickness
        self.shellAreaMap = {} # {partName:area}
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def initializeTable(self):
        table = []
        model = mdb.models[self.name]
        partNames = model.parts.keys()
        partNames.sort()
        for k in range(len(partNames)):
            partName = partNames[k]
            part = model.parts[partName]
            if len(part.cells) == 0 and len(part.faces) == 0:
                continue # Skip orphan mesh parts or line-based parts
            if not self.isInstanceDependent(model, partName):
                continue # Skip parts with independent instances

            table.append(getDefaultPartSettings(model, partName))
        return table
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSettings(self, table):
        settings = {}
        for row in table:
            settings[row[fullColTypeToColIdMap[NAME]]] = copy.deepcopy(row)
        return settings

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateNewPartInfo(self):
        tablePartNames = []
        for row in self.table:
            tablePartNames.append(row[fullColTypeToColIdMap[NAME]])
        model = mdb.models[self.name]
        a = model.rootAssembly
        session.customData.newPartInfo = RegisteredList()
        if len(model.parts.keys()) != len(tablePartNames):
            for partName in model.parts.keys():
                if partName not in tablePartNames:
                    p = model.parts[partName]
                    if len(p.cells) == 0 and len(p.faces) == 0: continue
                    if not self.isInstanceDependent(model, partName):
                        continue # Skip parts with independent instances
                    if partName[-3:] != partSuffix:
                        self.activationMap[partName] = True
                        row = getDefaultPartSettings(model, partName)
                        if row[fullColTypeToColIdMap[TECHNIQUE]] in ("Shell Sweep", "Solid Structured", "Solid Free"):
                            self.shellAreaMap[partName] = row[fullColTypeToColIdMap[VOLUME]]
                        session.customData.newPartInfo.append(row)
                        foundPart=False; isActive = False
                        for instanceName,instance in a.instances.items():
                            if instance.part.name == partName:
                                foundPart = True
                                if not a.features[instanceName].isSuppressed():
                                    isActive = True
                                    break
                        if foundPart:
                            self.activationMap[partName] = isActive

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateActivationMap(self, ):
        activationMap = RegisteredDictionary()
        isActive = {}; foundPart = {}
        model = mdb.models[self.name]
        a = model.rootAssembly
        for partName in model.parts.keys():
            activationMap[partName] = True
            isActive[partName] = False
            foundPart[partName] = False
        for instanceName,instance in a.instances.items():
            foundPart[instance.part.name] = True
            if not a.features[instanceName].isSuppressed():
                isActive[instance.part.name] = True
        for partName, activeFlag in isActive.items():
            if foundPart[partName]:
                activationMap[partName] = activeFlag
        return activationMap

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setValues(self, table, analysisType=EXPLICIT):
        self.table = table
        self.analysisType = analysisType
        self.generateMeshes()

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def generateMeshes(self):

        model = mdb.models[self.name]
        a = model.rootAssembly
        newTable = []
        for krow in range(len(self.table)):
            row = self.table[krow]
            status = row[fullColTypeToColIdMap[STATUS]]
            partName = row[fullColTypeToColIdMap[NAME]]

            # If (somehow) part is not in database
            # skip it
            #
            if partName not in model.parts.keys(): continue
            
            # Check if part instance is dependent
            #
            if not self.isInstanceDependent(model, partName):
                print 'Part "%s" has an independent instance. Skipped' % partName
                continue

            hasChanged = self.hasChanged(partName, row)
            if partName not in self.activationMap.keys():
                self.activationMap[partName] = True
            if self.activationMap[partName] == False:
                if partName in self.prevSettings.keys():
                    if hasChanged:
                        status = "N"
                    else:
                        status = self.prevSettings[partName][fullColTypeToColIdMap[STATUS]]
                else:
                    status = ""
                row = (status,) + row[1:]
                newTable.append(row)
                continue
            technique = row[fullColTypeToColIdMap[TECHNIQUE]]
            elemShape = row[fullColTypeToColIdMap[ELEMSHAPE]]
            generationMethod = row[fullColTypeToColIdMap[GENERATIONMETHOD]]
            geometricOrder = row[fullColTypeToColIdMap[GEOMETRICORDER]]
            formulation = row[fullColTypeToColIdMap[FORMULATION]]
            elemSize = row[fullColTypeToColIdMap[ELEMSIZE]]
            materialName = row[fullColTypeToColIdMap[MATERIAL]]
            errors = row[fullColTypeToColIdMap[ERRORS]]
            warnings = row[fullColTypeToColIdMap[WARNINGS]]
            makeRigid = row[fullColTypeToColIdMap[RIGID]]
            includeSkin = row[fullColTypeToColIdMap[SKIN]]
            thickness = row[fullColTypeToColIdMap[THICKNESS]]
            volume = row[fullColTypeToColIdMap[VOLUME]]
            p = model.parts[partName]
            if status == "Skip":
                if partName in self.prevSettings.keys():
                    status = self.prevSettings[partName][0]
                else:
                    status = ""
                row = (status,) + row[1:]
                newTable.append(row)
                continue
            if status == "Y" and hasChanged==False:
                newTable.append(row)
                continue
            if self.canSkipMeshing(partName, row):
                print 'Setting sections properties for part "%s".' % partName
                # Check only things that don't require rebuilding geometry or remeshing
                oldSettings = self.prevSettings[partName]
                oldMakeRigid = oldSettings[fullColTypeToColIdMap[RIGID]]
                try:

                    # Modify material/section assignments 
                    if generationMethod == "Extrude" or (technique in ("Solid Free", "Solid Structured", "Solid Sweep") and elemShape == "Quad Shell"):
                        p = model.parts[partName + partSuffix]
                    self.updateFromAssignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)

                    # Modify rigid Body Assignments
                    if makeRigid != oldMakeRigid:
                        self.assignRigidBodyIfNecessary(model, p, makeRigid, elemShape, generationMethod)
                    row = ("Y",) + row[1:]
                except:
                    print sys.exc_value
                    row = ("N",) + row[1:]
                status, percentFailed, percentWarning = self.hasValidMesh(model, p)
                if status:
                    status = "Y" # Has Mesh
                else: # No Mesh
                    status = "N" # Tried to Generate Mesh
                row = (status,) + row[1:]
                newTable.append(row)
                continue
            print 'Meshing and setting properties for part "%s".' % partName
            try:
                if generationMethod == "Extrude":
                    # Initially generate as a shell, then extrude to continuum shell
                    tmpElemShape = "Quad Shell"
                    self.updateFromModifyPart(model, partName, tmpElemShape, technique, generationMethod)
                    self.updateFromAssignElemSize(p, elemSize)
                    oldPartName = p.name
                    newPartName = oldPartName + partSuffix
                    p.generateMesh()
                    p = self.getMeshPart(model, p, elemSize)
                    self.removeTemporaryFeatures(model, oldPartName)
                    self.updateFromAssignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)
                    self.updateFromAssignElemShape(model, p, technique, elemShape, geometricOrder, formulation, generationMethod)
                    self.updateFromAssignSkin(model, p, includeSkin, makeRigid, materialName, elemShape)
                    self.replaceInstances(model, oldPartName, newPartName)
                    self.assignRigidBodyIfNecessary(model, p, makeRigid, elemShape, generationMethod)
                elif generationMethod == "Automatic":
                    # Only available for tet elements
                    oldPartName = p.name
                    self.removeOldMpParts(model, partName)
                    self.removeOldEntities(model, p)
                    self.updateFromAssignElemShape(model, p, technique, elemShape, geometricOrder, formulation, generationMethod)
                    elemShape, geometricOrder, formulation, elemSize = self.automaticallyMesh(model, p, geometricOrder, formulation, elemSize)
                    self.updateFromAssignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)
                    self.updateFromAssignSkin(model, p, includeSkin, makeRigid, materialName, elemShape)
                    self.assignRigidBodyIfNecessary(model, p, makeRigid, elemShape, generationMethod)
                elif technique in ("Solid Free", "Solid Structured", "Solid Sweep") and elemShape == "Quad Shell":
                    self.updateFromModifyPart(model, partName, elemShape, technique, generationMethod)
                    oldPartName = p.name
                    newPartName = oldPartName + partSuffix
                    p = model.Part(name=newPartName, objectToCopy=p)
                    self.removeTemporaryFeatures(model, oldPartName)
                    self.updateFromAssignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)
                    self.updateFromAssignElemShape(model, p, technique, elemShape, geometricOrder, formulation, generationMethod)
                    self.updateFromAssignElemSize(p, elemSize)
                    self.updateFromAssignSkin(model, p, includeSkin, makeRigid, materialName, elemShape)
                    self.replaceInstances(model, oldPartName, newPartName)
                    self.assignRigidBodyIfNecessary(model, p, makeRigid, elemShape, generationMethod)
                    p.generateMesh()
                else:
                    p = model.parts[partName]
                    self.removeOldMpParts(model, partName)
                    self.updateFromModifyPart(model, partName, elemShape, technique, generationMethod)
                    self.updateFromAssignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)
                    self.updateFromAssignElemShape(model, p, technique, elemShape, geometricOrder, formulation, generationMethod)
                    self.updateFromAssignElemSize(p, elemSize)
                    self.updateFromAssignSkin(model, p, includeSkin, makeRigid, materialName, elemShape)
                    self.assignRigidBodyIfNecessary(model, p, makeRigid, elemShape, generationMethod)
                    p.generateMesh()
            except:
                print sys.exc_value
                row = ("N",) + row[1:]
                newTable.append( row )
                continue

            # Check mesh and set status
            status, percentFailed, percentWarning = self.hasValidMesh(model, p)
            if status:
                status = "Y" # Has Mesh
            else:
                status = "N" # Tried to Generate Mesh, No mesh
            row = [status,] + list(row[1:])
            if generationMethod == "Automatic":
                row[fullColTypeToColIdMap[ELEMSHAPE]] = elemShape
                row[fullColTypeToColIdMap[GEOMETRICORDER]] = geometricOrder
                row[fullColTypeToColIdMap[FORMULATION]] = formulation
                row[fullColTypeToColIdMap[ELEMSIZE]] = elemSize
            row[fullColTypeToColIdMap[ERRORS]] = percentFailed
            row[fullColTypeToColIdMap[WARNINGS]] = percentWarning
            row = tuple(row)
            newTable.append(row)
        a.regenerate()            
        self.table = tuple(newTable)
        self.prevSettings = self.getSettings(self.table)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def removeOldMpParts(self, model, partName):
        mpName = partName + partSuffix
        if mpName in model.parts.keys():
            self.replaceInstances(model, mpName, partName)
            del model.parts[mpName]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateFromModifyPart(self, model, partName, elemShape, technique, generationMethod):
        try:
            self.modifyPart(model, partName, technique, elemShape, generationMethod)
        except:
            msg = self.getErrorMessage("while modifying part %s." % partName)
            raise AbaqusException, msg
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateFromAssignSectionAndMaterial(self, model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness):
        try:
            self.assignSectionAndMaterial(model, p, materialName, technique, elemShape, generationMethod, geometricOrder, formulation, thickness)
        except:
            msg = self.getErrorMessage("while assigning material properties to part %s." % p.name)
            raise AbaqusException, msg
            
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateFromAssignElemShape(self, model, p, technique, elemShape, geometricOrder, formulation, generationMethod):
        try:
            #if generationMethod == "Extrude":
            #    self.assignElemShape(p, "Quad Shell", generationMethod)
            #else:
            self.assignElemShape(p, technique, elemShape, geometricOrder, formulation, generationMethod)
        except:
            msg = self.getErrorMessage("while assigning element type to part %s." % p.name)
            raise AbaqusException, msg
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateFromAssignElemSize(self, p, elemSize):
        try:
            self.assignElemSize(p, elemSize)
        except:
            msg = self.getErrorMessage("while assigning element size to part %s." % p.name)
            raise AbaqusException, msg
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def updateFromAssignSkin(self, model, p, includeSkin, makeRigid, materialName, elemShape):
        if makeRigid: return
        try:
            self.assignSkinIfNecessary(model, p, includeSkin, materialName, elemShape)
        except:
            msg = self.getErrorMessage("while assigning skin properties to part %s." % p.name)
            raise AbaqusException, msg

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def isInstanceDependent(self, model, partName):
        a = model.rootAssembly
        part = model.parts[partName]
        for instanceName in a.instances.keys():
            instance = a.instances[instanceName]
            if not hasattr(instance, 'partName'): continue
            if instance.partName == partName:
                return instance.dependent

        # Instance does not exist, go ahead an mesh it
        return True

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def replaceInstances(self, model, oldPartName, newPartName):
        a = model.rootAssembly
        newPart = model.parts[newPartName]
        for instanceName in a.instances.keys():
            instance = a.instances[instanceName]
            if not hasattr(instance, 'partName'): continue
            if instance.partName == oldPartName:
                instance.replace(instanceOf=newPart) 
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getErrorMessage(self, msg):
        errorMsg = "\n%s: %s. The error occured %s Mesh creation for this part will be skipped.\n\n" % \
                   (sys.exc_type, sys.exc_value, msg)
        return errorMsg

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def modifyPart(self, model, partName, technique, elemShape, generationMethod):
        p = model.parts[partName]
        self.removeOldEntities(model, p)
        self.assignVirtualTopologyIfNecessary(model, p, generationMethod)
        if generationMethod == "Extrude":
            self.modifySolidPartToBottomShellPart(model, p)
        elif technique in ("Solid Free", "Solid Structured", "Solid Sweep") and elemShape == "Quad Shell":
            self.modifySolidPartToMidSurfaceShellPart(model, p)
            
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignVirtualTopologyIfNecessary(self, model, p, generationMethod, useDefaults=True):
        # Virtual Topology
        if p.features.has_key("PME_VirtualTopology"):
            del p.features["PME_VirtualTopology"]
        if generationMethod == "Virtual Topology":
            if useDefaults:
                shortEdgeThreshold=0.16
                smallFaceAreaThreshold=0.13
                faceAspectRatioThreshold=10.0
                smallFaceCornerAngleThreshold=10.0
                thinStairFaceThreshold=0.032
                cornerAngleTolerance=30.0
                blendSubtendedAngleTolerance=60.0
                blendRadiusTolerance=0.8
            else:
                # Find max and min edge lengths
                maxEdgeLength = 0.0
                for edge in p.edges:
                    edgeLength = edge.getSize(printResults=False)
                    maxEdgeLength = max(maxEdgeLength, edgeLength)
                # Find max and min face sizes
                maxFaceArea = 0.0
                for face in p.faces:
                    faceArea = face.getSize(printResults=False)
                    maxFaceArea = max(maxFaceArea, faceArea)
                shortEdgeThreshold = 1.0*maxEdgeLength,
                smallFaceAreaThreshold = 1.0*maxFaceArea
                faceAspectRatioThreshold = 10.0
                thinStairFaceThreshold = 0.025*maxEdgeLength
                cornerAngleTolerance = 45.0,
                blendSubtendedAngleTolerance = 60.0
                blendRadiusTolerance = 0.1*maxEdgeLength
            try:
                f = p.createVirtualTopology(shortEdgeThreshold = shortEdgeThreshold,
                                            smallFaceAreaThreshold = smallFaceAreaThreshold,
                                            faceAspectRatioThreshold = faceAspectRatioThreshold,
                                            smallFaceCornerAngleThreshold = smallFaceCornerAngleThreshold,
                                            thinStairFaceThreshold = thinStairFaceThreshold,
                                            cornerAngleTolerance = cornerAngleTolerance,
                                            blendSubtendedAngleTolerance = blendSubtendedAngleTolerance,
                                            blendRadiusTolerance = blendRadiusTolerance,
                                            mergeShortEdges=True, mergeSmallFaces=True, mergeSliverFaces=True, applyBlendControls=True,
                                            mergeSmallAngleFaces=True, mergeThinStairFaces=True, ignoreRedundantEntities=True,
                                            )
                lastFeatureName = p.features.keys()[-1]
                if "Auto virtual" in lastFeatureName:
                    p.features.changeKey(fromName=p.features.keys()[-1], toName="PME_VirtualTopology")
            except:
                print 'Warning: %s Assignment of virtual topology failed on part "%s".' % (sys.exc_value, p.name)
         
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getMeshPart(self, model, p_orig, elemSize):
        thickness = self.shellThicknessesFromSolids[p_orig.name]
        newName = p_orig.name + partSuffix
        p_orig.PartFromMesh(name=newName)
        p = model.parts[newName]
        numLayers = max(1, int(thickness/elemSize))
        p.generateMeshByOffset(region=regionToolset.Region(side1Elements=p.elements[:]),
                               meshType=SOLID, totalThickness=thickness, 
                               numLayers=numLayers, offsetDirection=OUTWARD, deleteBaseElements=TRUE)
        return p

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def removeTemporaryFeatures(self, model, partName):
        if partName not in model.parts.keys(): return
        p = model.parts[partName]
        featureNames =  p.features.keys()
        featureNames.reverse()
        for featureName in featureNames:
            if featureName in ("PME_RemoveFaces", "PME_AssignMidsurface", "PME_MidsurfaceOffset"):
                del p.features[featureName]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def removeOldEntities(self, model, p):
        a = model.rootAssembly
        # Remove old shell mesh parts
        
        # Remove old section assignments
        del p.sectionAssignments[:]

        # Remove old mesh partitions/removed faces
        self.removeTemporaryFeatures(model, p.name)
    
        # Remove old reinforcements (skins)
        for skinName in p.skins.keys():
            if skinName == 'PME_OuterSkin':
                del p.skins[skinName]

        # Remove old top surface
        if p.surfaces.has_key('PME_TopSurf'):
            del p.surfaces['PME_TopSurf']

        # Remove reference point
        if p.features.has_key("PME_RigidBodyRefPt"):
            del p.features["PME_RigidBodyRefPt"]

        # Remove rigid body definition
        mpPartName = p.name + partSuffix
        for instanceName, instance in a.instances.items():
            if not hasattr(instance, 'partName'): continue
            if instance.partName in (p.name, mpPartName):
                constraintName = "PME_%s_rb" % instanceName
                if constraintName in model.constraints.keys():
                    del model.constraints[constraintName]        
        p.regenerate()
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignElemShape(self, p, technique, elemShape, geometricOrder, formulation, generationMethod):
        # set default element types
        if formulation == "Continuum Shell":
            # Hack: Have to first assign elements as HEX, then CONT_SHELL (CAE BUG)
            #elemType1 = mesh.ElemType(elemCode=C3D8R, elemLibrary=self.analysisType) 
            #elemType2 = mesh.ElemType(elemCode=C3D6, elemLibrary=self.analysisType)
            #elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
            #elemTypes = (elemType1, elemType2, elemType3)
            #p.setElementType(regions=p.sets['PME_Body'], elemTypes=elemTypes)
            p.setMeshControls(regions=p.cells[:], elemShape=HEX,  technique=SWEEP, allowMapped=True)
            elemType1 = mesh.ElemType(elemCode=SC8R, elemLibrary=self.analysisType)
            elemType2 = mesh.ElemType(elemCode=SC6R, elemLibrary=self.analysisType)
            elemType3 = mesh.ElemType(elemCode=UNKNOWN_TET, elemLibrary=self.analysisType)
            elemTypes = (elemType1, elemType2)
            p.setElementType(regions=p.sets['PME_Body'], elemTypes=elemTypes)
        elif elemShape == "Hex":            
            if geometricOrder == "Linear":
                if formulation == "Reduced Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D8R, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
                elif formulation == "Reduced Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D8RH, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
                elif formulation == "Full Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D8, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
                elif formulation == "Full Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D8H, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4H, elemLibrary=self.analysisType)
                elif formulation == "Incompatible Modes":
                    elemType1 = mesh.ElemType(elemCode=C3D8I, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
                elif formulation == "Incompatible Modes, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D8IH, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4H, elemLibrary=self.analysisType)
                elif formulation == "Continuum Shell":
                    elemType1 = mesh.ElemType(elemCode=SC8R, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=SC6R, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
            else:
                if formulation == "Reduced Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D20R, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10, elemLibrary=self.analysisType)
                elif formulation == "Reduced Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D20RH, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10H, elemLibrary=self.analysisType)
                elif formulation == "Full Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D20, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10, elemLibrary=self.analysisType)
                elif formulation == "Full Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D20H, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10H, elemLibrary=self.analysisType)
            elemTypes = (elemType1, elemType2, elemType3)
            p.setElementType(regions=p.sets['PME_Body'], elemTypes=elemTypes)
            if generationMethod != "Extrude":
                p.setMeshControls(regions=p.cells[:], elemShape=HEX, technique=SYSTEM_ASSIGN)
        elif elemShape == "Tet":
            if geometricOrder == "Linear":
                if formulation == "Full Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D8, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4, elemLibrary=self.analysisType)
                elif formulation == "Full Integration, Hybrid": 
                    elemType1 = mesh.ElemType(elemCode=C3D8H, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D6H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D4H, elemLibrary=self.analysisType)
            else:
                if formulation == "Full Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D20, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10, elemLibrary=self.analysisType)
                elif formulation == "Modified Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D20, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10M, elemLibrary=self.analysisType)
                elif formulation == "Full Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D20H, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10H, elemLibrary=self.analysisType)
                elif formulation == "Modified Integration, Hybrid":
                    elemType1 = mesh.ElemType(elemCode=C3D20H, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15H, elemLibrary=self.analysisType)
                    elemType3 = mesh.ElemType(elemCode=C3D10MH, elemLibrary=self.analysisType)
                elif formulation == "Improved Surface Integration":
                    elemType1 = mesh.ElemType(elemCode=C3D20, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=C3D15, elemLibrary=self.analysisType)                  
                    elemType3 = mesh.ElemType(elemCode=C3D10I, elemLibrary=self.analysisType)
            elemTypes = (elemType1, elemType2, elemType3)
            p.setElementType(regions=p.sets['PME_Body'], elemTypes=elemTypes)
            p.setMeshControls(regions=p.cells[:], elemShape=TET, technique=FREE, allowMapped=True)
        elif elemShape == "Quad Shell":
            if geometricOrder == "Linear":
                if formulation == "Full Integration":
                    elemType1 = mesh.ElemType(elemCode=S4, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=S3, elemLibrary=self.analysisType)
                elif formulation == "Reduced Integration":
                    elemType1 = mesh.ElemType(elemCode=S4R, elemLibrary=self.analysisType)
                    elemType2 = mesh.ElemType(elemCode=S3, elemLibrary=self.analysisType)
            else:
                elemType1 = mesh.ElemType(elemCode=S8R, elemLibrary=self.analysisType)
                elemType2 = mesh.ElemType(elemCode=STRI65, elemLibrary=self.analysisType)
            elemTypes = (elemType1, elemType2)
            p.setElementType(regions=p.sets['PME_Body'], elemTypes=elemTypes)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignSectionAndMaterial(self, model, p, materialName, technique, elemShape, generationMethod,  geometricOrder, formulation, thickness):
        if elemShape == "Quad Shell" and technique not in ("Shell Free", "Shell Structured", "Shell Sweep") or formulation == "Continuum Shell":
            if p.name[-3:] == partSuffix:
                thickness = self.shellThicknessesFromSolids[p.name[:-3]]
            else:
                thickness = self.shellThicknessesFromSolids[p.name]
        del p.sectionAssignments[:]
        if technique in ("Shell Free", "Shell Structured", "Shell Sweep"):
            p.Set(name='PME_Body', faces=p.faces[:])
            sectionName = "PME_Shell_%s" % materialName
            if sectionName not in model.sections.keys():
                model.HomogeneousShellSection(name=sectionName, material=materialName, thickness=thickness)
        elif formulation == "Continuum Shell":
            if generationMethod=="Extrude":
                p.Set(name='PME_Body', elements=p.elements[:])
                numThick = self.getThickNumber(materialName, thickness)
                sectionName = "PME_Shell_%s_%s" % (materialName,numThick)
                if sectionName not in model.sections.keys():
                    model.HomogeneousShellSection(name=sectionName, material=materialName, thickness=thickness)
            else:
                p.Set(name='PME_Body', cells=p.cells[:])
                sectionName = "PME_Solid_%s" % materialName
                if sectionName not in model.sections.keys():
                    model.HomogeneousSolidSection(name=sectionName, material=materialName, thickness=1.0)
        elif elemShape in ("Tet","Hex"):
            if generationMethod=="Extrude":
                p.Set(name='PME_Body', elements=p.elements[:])
            else:
                p.Set(name='PME_Body', cells=p.cells[:])                
            sectionName = "PME_Solid_%s" % materialName
            if sectionName not in model.sections.keys():
                model.HomogeneousSolidSection(name=sectionName, material=materialName, thickness=1.0)
        elif elemShape in "Quad Shell":
            p.Set(name='PME_Body', faces=p.faces[:])
            numThick = self.getThickNumber(materialName, thickness)
            sectionName = "PME_Shell_%s_%s" % (materialName,numThick)
            if sectionName not in model.sections.keys():
                model.HomogeneousShellSection(name=sectionName, material=materialName, thickness=thickness)
        p.SectionAssignment(region=p.sets['PME_Body'], sectionName=sectionName)
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getThickNumber(self, materialName, thickness):
        if materialName not in self.shellThicknessesToMaterialMap.keys():
            self.shellThicknessesToMaterialMap[materialName] = {}
        for thick in self.shellThicknessesToMaterialMap[materialName].keys():
            if self.isWithinTolerance(thick, thickness):
                return self.shellThicknessesToMaterialMap[materialName][thick]
        numThicks = self.shellThicknessesToMaterialMap[materialName].values()
        numThicks.sort()
        if len(numThicks) > 0:
            numThick = numThicks[-1] + 1
        else:
            numThick = 1
        self.shellThicknessesToMaterialMap[materialName][thickness] = numThick
        return numThick
         
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignSkinIfNecessary(self, model, p, includeSkin, materialName, elemShape):
        if 'PME_OuterSkin' in p.skins.keys():
            del p.skins['PME_OuterSkin']
        if not includeSkin:
            if p.sets.has_key('PME_OuterSkin'):
                del p.sets['PME_OuterSkin']
            if p.surfaces.has_key('PME_OuterSkin'):
                del p.surfaces['PME_OuterSkin']
            return
        if elemShape == "Quad Shell": return
        # Generate Section
        sectionName = "PME_Skin_%s" % materialName
        if sectionName not in model.sections.keys():
            model.MembraneSection(name=sectionName, material=materialName, 
                                  thickness=1.E-6, poissonDefinition=DEFAULT)
        # Generate Skin Surface
        if len(p.faces) > 0:
            if not p.surfaces.has_key('PME_OuterSkin') or len(p.surfaces['PME_OuterSkin'].faces)==0 or \
                   not p.sets.has_key('PME_OuterSkin') or len(p.sets['PME_OuterSkin'].faces)==0:
                # Geometry-based part
                faces = p.faces[:]
                for face in p.faces:
                    if len(face.getCells()) == 1:
                        faces = faces + p.faces[face.index:face.index+1]
                p.Surface(name='PME_OuterSkin', side1Faces=faces)
                p.Set(faces=faces, reinforcement=True, name='PME_OuterSkin')
            p.Skin(name='PME_OuterSkin', faces=self.getRegionSequenceFromSurface(p,'PME_OuterSkin'))  
        elif len(p.cells) == 0:
            if not p.surfaces.has_key('PME_OuterSkin'):
                # Orphan mesh part (creates set, surface, and skin)
                self.generateOrphanMeshSkinSurface(p)
        elemType1 = mesh.ElemType(elemCode=M3D4R, elemLibrary=self.analysisType) 
        elemType2 = mesh.ElemType(elemCode=M3D3, elemLibrary=self.analysisType)
        p.setElementType(regions=p.sets['PME_OuterSkin'], elemTypes=(elemType1, elemType2))
        p.SectionAssignment(region=p.sets['PME_OuterSkin'], sectionName=sectionName, offset=0.0,
                            offsetType=MIDDLE_SURFACE, offsetField='',thicknessAssignment=FROM_SECTION)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def generateOrphanMeshSkinSurface(self, p):
        # Only used for continuum shell mesh parts
        args = {}
        args['name'] = "PME_OuterSkin"
        face1Ids = [] 
        face2Ids = []
        face3Ids = []
        face4Ids = []
        face5Ids = []
        face6Ids = []
        pElementFaces = p.elementFaces
        skinFaces = pElementFaces[0:0]
        for k in range(len(pElementFaces)):
            elemFace = pElementFaces[k]
            elements = elemFace.getElements()
            if len(elements) != 1: continue
            skinFaces = skinFaces + pElementFaces[k:k+1]
            if elemFace.face == FACE1:
                face1Ids.append(elements[0].label)
            if elemFace.face == FACE2:
                face2Ids.append(elements[0].label)
            if elemFace.face == FACE3:
                face3Ids.append(elements[0].label)
            if elemFace.face == FACE4:
                face4Ids.append(elements[0].label)
            if elemFace.face == FACE5:
                face5Ids.append(elements[0].label)
            if elemFace.face == FACE6:
                face6Ids.append(elements[0].label)
        if len(face1Ids) > 0:
            args['face1Elements'] = p.elements.sequenceFromLabels(face1Ids)
        if len(face2Ids) > 0:
            args['face2Elements'] =  p.elements.sequenceFromLabels(face2Ids)
        if len(face3Ids) > 0:
            args['face3Elements'] =  p.elements.sequenceFromLabels(face3Ids)
        if len(face4Ids) > 0:
            args['face4Elements'] =  p.elements.sequenceFromLabels(face4Ids)
        if len(face5Ids) > 0:
            args['face5Elements'] =  p.elements.sequenceFromLabels(face5Ids)
        if len(face6Ids) > 0:
            args['face6Elements'] =  p.elements.sequenceFromLabels(face6Ids)
        apply(p.Surface, (), args)
        p.Skin(name='PME_OuterSkin', elementFaces=skinFaces)
        p.Set(elements=p.skins['PME_OuterSkin'].elements[:], name='PME_OuterSkin')

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def removeOldRigidDefinitions(self, model, p):
        a = model.rootAssembly
        for instanceName, instance in a.instances.items():
            if not hasattr(instance, 'partName'): continue
            if instance.partName == p.name:
                constraintName = "PME_%s_rb" % instanceName
                if constraintName in model.constraints.keys():
                    del model.constraints[constraintName]
                if p.features.has_key("PME_RigidBodyRefPt"):
                    del p.features["PME_RigidBodyRefPt"]

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignRigidBodyIfNecessary(self, model, p, makeRigid, elemShape, generationMethod):
        if not makeRigid:
            self.removeOldRigidDefinitions(model, p)
            return
        a = model.rootAssembly
        # Create reference point
        if generationMethod == "Extrude":
            # Is there a better way?
            xsum=0.0; ysum=0.0; zsum=0.0
            for node in p.nodes:
                coords = node.coordinates
                xsum = xsum + coords[0]
                ysum = ysum + coords[1]
                zsum = zsum + coords[2]
            xcent = xsum/float(len(p.nodes))
            ycent = ysum/float(len(p.nodes))
            zcent = zsum/float(len(p.nodes))
            cen = (xcent, ycent, zcent)
        if elemShape == "Quad Shell":
            cen = p.getCentroid(faces=p.faces[:])
        else:
            cen = p.getCentroid(cells=p.cells[:])

        if p.features.has_key("PME_RigidBodyRefPt"):
            del p.features["PME_RigidBodyRefPt"]
        f = p.ReferencePoint(point=cen)
        p.features.changeKey(fromName=f.name, toName="PME_RigidBodyRefPt")
        refPtId = f.id
        
        # Create Rigid Body Definitions
        if a.isOutOfDate:
            a.regenerate()
        a = model.rootAssembly
        for instanceName,instance in a.instances.items():
            if not hasattr(instance, 'partName'): continue
            if instance.partName == p.name:
                constraintName = "PME_%s_rb" % instanceName
                bodyRegion = instance.sets["PME_Body"]
                refPointRegion = regionToolset.Region(referencePoints=(instance.referencePoints[refPtId],))
                model.RigidBody(name=constraintName, refPointRegion=refPointRegion,
                                bodyRegion=bodyRegion, refPointAtCOM=ON)
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def assignElemSize(self, p, elemSize):
        p.seedPart(size=elemSize, deviationFactor=0.1)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def modifySolidPartToMidSurfaceShellPart(self, model, p):
        if not p.surfaces.has_key("PME_BotSurf"):
            msg = "WARNING: Part '%s' does not have a face 'PME_BotSurf' defining the surface." \
                  "The part will not be meshed." % p.name         
            raise AbaqusException, msg
        # Create Top Face
        oppositeFaceIds = self.getFaceIdsOfOppositeSurface(p, "PME_BotSurf")
        if len(oppositeFaceIds) == 0:
            msg = "Could not find a surface opposite and idential to 'PME_BotSurf'"
            raise AbaqusException, msg
        elif len(oppositeFaceIds) != len(p.surfaces["PME_BotSurf"].faces):
            msg = "WARNING: Could not find a surface opposite and idential to 'PME_BotSurf'\n" \
                  "Surface 'PME_TopSurf' has a different number of faces. This could cause \n" \
                  "problems with mesh creation.\n\n"
            print msg
        p.Surface(name="PME_TopSurf", side1Faces=self.getSequenceFromIds(p, "faces", oppositeFaceIds))
        self.shellThicknessesFromSolids[p.name] = self.getThicknessFromTopAndBotSurf(p)
        
        # Remove cells, leave BotSurf shell
        pointOns = []
        for face in p.surfaces['PME_BotSurf'].faces:
            pointOns.append(face.pointOn[0])
        f = p.AssignMidsurfaceRegion(cellList = p.cells[:])
        p.features.changeKey(fromName=f.name, toName="PME_AssignMidsurface")
        botFaces = self.getRegionSequenceFromPointOns(p, "faces", pointOns)
        f = p.OffsetFaces(faceList = botFaces, distance=self.shellThicknessesFromSolids[p.name]/2.0, trimToReferenceRep=False)
        p.features.changeKey(fromName=f.name, toName="PME_MidsurfaceOffset")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def modifySolidPartToBottomShellPart(self, model, p):
        if not p.surfaces.has_key("PME_BotSurf"):
            msg = "WARNING: Part '%s' does not have a face 'PME_BotSurf' defining the surface." \
                  "The part will not be meshed." % p.name         
            raise AbaqusException, msg
        # Create Top Face
        oppositeFaceIds = self.getFaceIdsOfOppositeSurface(p, "PME_BotSurf")
        if len(oppositeFaceIds) == 0:
            msg = "Could not find a surface opposite and idential to 'PME_BotSurf'"
            raise AbaqusException, msg
        elif len(oppositeFaceIds) != len(p.surfaces["PME_BotSurf"].faces):
            msg = "**WARNING: Could not find a surface opposite and idential to 'PME_BotSurf'\n" \
                  "Surface 'PME_TopSurf' has a different number of faces. This could cause \n" \
                  "problems with mesh creation.\n\n"
            print msg
        p.Surface(name="PME_TopSurf", side1Faces=self.getSequenceFromIds(p, "faces", oppositeFaceIds))
        self.shellThicknessesFromSolids[p.name] = self.getThicknessFromTopAndBotSurf(p)
    
        # Remove cells, leave BotSurf shell
        botSurfIds = []
        for face in p.surfaces["PME_BotSurf"].faces:
            botSurfIds.append(face.index)
        facesToRemove = None
        for face in p.faces:
            if face.index not in botSurfIds:
                if facesToRemove:
                    facesToRemove = facesToRemove + p.faces[face.index:face.index+1]
                else:
                    facesToRemove = p.faces[face.index:face.index+1]
        f = p.RemoveFaces(faceList=facesToRemove, deleteCells=TRUE)
        p.features.changeKey(fromName=f.name, toName="PME_RemoveFaces")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getThicknessFromTopAndBotSurf(self, p):
        # Must be called before solid part is converted to shell
        pEdges = p.edges
        # Determine Thickness
        numThicks = 0
        thickSum = 0.0
        oppositeFaceIds =  self.getRegionIdsFromRegion(p, "surfaces", "PME_TopSurf", "faces")
        connEdgeIds = []
        for face in p.surfaces["PME_BotSurf"].faces:
            edgeIds = self.getConnectingEdgeIds(p, face.index)
            for edgeId in edgeIds:
                if edgeId not in connEdgeIds:
                    connEdgeIds.append(edgeId)
                    numThicks = numThicks + 1
                    thickSum = thickSum + pEdges[edgeId].getSize(printResults=False)
        thickness = thickSum/float(numThicks)
        return thickness

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def hasChanged(self, partName, newSettings):
        if partName not in self.prevSettings.keys(): return False
        oldSettings = self.prevSettings[partName]
        for k in range(2,len(newSettings)-1):
            newValue =  newSettings[k]
            oldValue = oldSettings[k]
            try:
                newValue = float(newValue)
                oldValue = float(oldValue)
                if not self.isWithinTolerance(newValue, oldValue, 1.E-6):
                    return True
            except:
                if newSettings[k] != oldSettings[k]:
                    return True
        return False

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def canSkipMeshing(self, partName, newSettings):
        if partName not in self.prevSettings.keys(): return False
        oldSettings = self.prevSettings[partName]
        status = newSettings[fullColTypeToColIdMap[STATUS]]
        technique = newSettings[fullColTypeToColIdMap[TECHNIQUE]]
        elemShape = newSettings[fullColTypeToColIdMap[ELEMSHAPE]]
        generationMethod = newSettings[fullColTypeToColIdMap[GENERATIONMETHOD]]
        geometricOrder = newSettings[fullColTypeToColIdMap[GEOMETRICORDER]]
        formulation = newSettings[fullColTypeToColIdMap[FORMULATION]]
        elemSize = newSettings[fullColTypeToColIdMap[ELEMSIZE]]
        skin =  newSettings[fullColTypeToColIdMap[SKIN]]
        oldStatus = oldSettings[fullColTypeToColIdMap[STATUS]]
        oldTechnique = oldSettings[fullColTypeToColIdMap[TECHNIQUE]]
        oldElemShape = oldSettings[fullColTypeToColIdMap[ELEMSHAPE]]
        oldGenerationMethod = oldSettings[fullColTypeToColIdMap[GENERATIONMETHOD]]
        oldGeometricOrder = oldSettings[fullColTypeToColIdMap[GEOMETRICORDER]]
        oldFormulation = oldSettings[fullColTypeToColIdMap[FORMULATION]]
        oldElemSize = oldSettings[fullColTypeToColIdMap[ELEMSIZE]]
        oldSkin = oldSettings[fullColTypeToColIdMap[SKIN]]
        if status == "Skip": return True
        if oldStatus == 'N': return False
        if oldStatus == '': return False
        if oldTechnique != technique: return False
        if oldElemShape != elemShape: return False
        if oldGenerationMethod != generationMethod: return False
        if oldGeometricOrder != geometricOrder: return False
        if oldFormulation != formulation: return False
        if oldElemSize != elemSize: return False
        if oldSkin != skin: return False
        model = mdb.models[self.name]
        status, percentFailed, percentWarning = self.hasValidMesh(model, model.parts[partName])
        return status

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def hasValidMesh(self, model, p):
        if len(p.elements) == 0:
            return False, 0.0, 0.0
        status = True
        if len(p.cells) > 0 and not p.features.has_key("PME_AssignMidsurface"):
            meshStats = p.getMeshStats(regions=p.cells[:])
            if meshStats.numMeshedRegions != len(p.cells):
                status = False
            meshQuality = p.verifyMeshQuality(criterion=ANALYSIS_CHECKS, regions=p.cells[:])
        elif len(p.faces) > 0:
            meshStats = p.getMeshStats(regions=p.faces[:])
            if len(p.cells) > 0:
                if meshStats.numMeshedRegions == 0:
                    status = False
            else:
                if meshStats.numMeshedRegions != len(p.faces):
                    status = False
            meshQuality = p.verifyMeshQuality(criterion=ANALYSIS_CHECKS, regions=p.faces[:])
        elif len(p.elements) > 0:
            meshQuality = p.verifyMeshQuality(criterion=ANALYSIS_CHECKS, regions=p.elements[:])
        else:
            return False, 0.0, 0.0
        percentFailed = 100.0*float(len(meshQuality["failedElements"]))/float(len(p.elements))
        percentWarning = 100.0*float(len(meshQuality["warningElements"]))/float(len(p.elements))
        if percentFailed != 0.0:
            status = False
        return status, percentFailed, percentWarning

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def resumeInstances(self, partNames):
        model = mdb.models[self.name]
        a = model.rootAssembly
        instanceNames = []
        for instance in a.instances.values():
            if instance.part.name in partNames:
                instanceNames.append(instance.name)
        a.resumeFeatures(instanceNames)
        for partName in partNames:
            self.activationMap[partName] = True

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def suppressInstances(self, partNames):
        model = mdb.models[self.name]
        a = model.rootAssembly
        instanceNames = []
        for instance in a.instances.values():
            if instance.part.name in partNames:
                instanceNames.append(instance.name)
        a.suppressFeatures(instanceNames)
        for partName in partNames:
            self.activationMap[partName] = False
        
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def isWithinTolerance(self, a, b, tol=1.E-4):
        if a == b:
            return 1
        if math.fabs(a) > math.fabs(b):
            denom = a
        else:
            denom = b
        diff = math.fabs( (a-b)/denom )
        if diff < tol:
            return 1
        else:
            return 0

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def automaticallyMesh(self, model, p, geometricOrder, formulation, elemSize):
        maxTry=5
        mapOn=False
        incSize=False
        applyVT=True
        refinePerLevel = 0.9
        elemShape = "Tet"
        if p.isLocked():
            try:
                p.Unlock(reportWarnings=False)
            except:
                print 'Failed to unlock part "%s". Skipping it' % p.name
                return elemShape, geometricOrder, formulation, elemSize
        # First try to refine mesh
        for refineLevel in range(maxTry):
            status, elemSize = self.setPartMeshControls(p, mapOn, refineLevel, refinePerLevel, elemSize)
            if status:
                try:
                    p.generateMesh()
                    status, percentFailed, percentWarning = self.hasValidMesh(model, p)
                except:
                    continue
                if status: return elemShape, geometricOrder, formulation, elemSize
            else:
                return elemShape, geometricOrder, formulation, elemSize
        # If mesh refinement doesn't work, try virtual topology
        self.assignVirtualTopologyIfNecessary(model, p, "Virtual Topology", useDefaults=False)
        return elemShape, geometricOrder, formulation, elemSize

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setPartMeshControls(self, p, mapOn, refineLevel, refinePerLevel, elemSize):
        origSize = elemSize
        try:
            p.setMeshControls(regions=p.cells[:], sizeGrowth=MODERATE, allowMapped=mapOn)
            if refineLevel > 0:
                elemSize = elemSize * pow(refinePerLevel, refineLevel)
                p.seedPart(size=elemSize)
        except:
                print '%s Could not assign seed size of "%s" to part "%s".' % (sys.exc_value, elemSize, p.name)
                return False, origSize
        return True, elemSize
