from abaqusGui import *
from abaqusConstants import *
from abq_Pme.pmeConstants import *
from abq_Pme.pmeModelDB import PmeModelDB, tableFeatures

shellFreeIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconShellFree.png'))
solidFreeIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidFree.png'))
solidSweptIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidSwept.png'))
solidStructuredIcon = afxCreatePNGIcon(os.path.join('abq_Pme','pmeIconSolidStructured.png'))
shellFreeIcon.create()
solidFreeIcon.create()
solidSweptIcon.create()
solidStructuredIcon.create()

###########################################################################
# Class definition
###########################################################################

class PmeModelForm(AFXForm):

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, owner):
        AFXForm.__init__(self, owner)
        self.cmd = AFXGuiCommand(self, "setValues", "_UNDEFINED", True)
  
        self.tableKw =  AFXTableKeyword(self.cmd, 'table', True)
        self.numShells = 0

        self.db = None
        self.tableKw.setColumnType(0, AFXTABLE_TYPE_STRING) # Status
        self.tableKw.setColumnType(1, AFXTABLE_TYPE_STRING) # Part Name
        self.tableKw.setColumnType(2, AFXTABLE_TYPE_STRING) # Technique
        self.tableKw.setColumnType(3, AFXTABLE_TYPE_STRING) # Element Shape
        self.tableKw.setColumnType(4, AFXTABLE_TYPE_STRING) # Method
        self.tableKw.setColumnType(5, AFXTABLE_TYPE_STRING) # Geometric Order
        self.tableKw.setColumnType(6, AFXTABLE_TYPE_STRING) # Formulation
        self.tableKw.setColumnType(7, AFXTABLE_TYPE_FLOAT) # Element size
        self.tableKw.setColumnType(8, AFXTABLE_TYPE_STRING) # Material Name
        self.tableKw.setColumnType(9, AFXTABLE_TYPE_FLOAT) # Percent Errors
        self.tableKw.setColumnType(10, AFXTABLE_TYPE_FLOAT) # Percent Warnings
        self.tableKw.setColumnType(11, AFXTABLE_TYPE_BOOL) # Make Rigid
        self.tableKw.setColumnType(12, AFXTABLE_TYPE_BOOL) # Include Skin
        self.tableKw.setColumnType(13, AFXTABLE_TYPE_FLOAT) # Thickness
        self.tableKw.setColumnType(14, AFXTABLE_TYPE_FLOAT) # Volume

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setPmeModelObj(self, modelName=None):
        
        from kernelAccess import mdb,session
        if modelName:
            self.modelName = modelName
        else:
            self.modelName = session.sessionState[session.currentViewportName]['modelName']
        if not hasattr(mdb.customData, 'pmeModels'):
            cmd = "abq_Pme.pmeModule.InitializeCustomData('%s')\n" % self.modelName
            sendCommand(cmd)

        elif not mdb.customData.pmeModels.has_key(self.modelName):
            cmd = "abq_Pme.pmeModule.InitializeCustomData('%s')\n" % self.modelName
            sendCommand(cmd)

        self.pmeModelObj = mdb.customData.pmeModels[self.modelName]
        self.pmeModelObjString = "mdb.customData.pmeModels['%s']" % self.modelName
        self.cmd.setObjectName(self.pmeModelObjString)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def activate(self):
        AFXForm.activate(self)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getFirstDialog(self):
        self.setPmeModelObj() # sets self.modelName from current viewport
        self.db = PmeModelDB(self)
        self.db.fillTableFromCustomData()
        return self.db

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def verifyKeywordValues(self):
        model = mdb.models[self.modelName]
        numCols = self.db.table.getNumColumns()
        # Clean up old tableKw
        if self.tableKw.getNumRows() >= self.db.table.getNumRows():
            self.tableKw.removeRows(self.db.table.getNumRows()-1, self.tableKw.getNumRows()-self.db.table.getNumRows()-1)
        # Loop over table
        for krow in range(1,self.db.table.getNumRows()):
            partName = self.db.getItemValue(krow, NAME)
            technique = self.db.getItemValue(krow, TECHNIQUE)
            for kcol in range(len(fullColIdToColTypeMap.keys())):
                colType = fullColIdToColTypeMap[kcol]
                value = self.db.getItemValue(krow, colType)
                # Perform Checks
                if colType == ELEMSIZE:
                    try:
                        elemSize = float(value)
                    except:
                        msg = "An invalid element size was entered for part '%s'." % partName
                        showAFXErrorDialog(getAFXApp().getAFXMainWindow(), msg)                
                        return 0
                elif colType == THICKNESS:
                    if technique in ("Shell Free", "Shell Structured", "Shell Sweep"):
                        try:
                            thickness =  float(value)
                        except:
                            msg = "An invalid shell thickness was entered for part '%s'." % partName
                            showAFXErrorDialog(getAFXApp().getAFXMainWindow(), msg)                
                            return 0
                elif colType == MATERIAL:
                    p = model.parts[partName]
                    # Check for valid material name
                    if value not in model.materials.keys():
                        msg = "Part '%s' has not been assigned a valid material name." % partName
                        showAFXErrorDialog(getAFXApp().getAFXMainWindow(), msg)                
                        return 0
                # Set value in tableKw
                if tableFeatures[colType][1] == AFXTable.BOOL:
                    if value:
                        self.tableKw.setValue(krow-1, kcol, "True")
                    else:
                        self.tableKw.setValue(krow-1, kcol, "False")
                else:
                    self.tableKw.setValue(krow-1, kcol, '%s' % value)
        # Set column types

        return 1

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def doCustomTasks(self):
        self.db.updateStatuses()


class PmeModelGenericForm(AFXForm):
    # Used for calls with no arguments
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, owner):
        AFXForm.__init__(self, owner)
        self.cmd = AFXGuiCommand(self, "_UNDEFINED", "_UNDEFINED", True)
        self.partNamesKw = AFXTupleKeyword(self.cmd, 'partNames', True)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def setValues(self, modelName, partNames, method):
        self.modelName = modelName
        self.partNamesKw.setMaxLength(len(partNames))
        for k in range(len(partNames)):
            self.partNamesKw.setElementType(k, AFXTUPLE_TYPE_STRING)
            self.partNamesKw.setValue(k,partNames[k])
        self.cmd.setObjectName("mdb.customData.pmeModels['%s']" % self.modelName)
        self.cmd.setMethod(method)
