# Routines for manipulating part geometry - assumes extruded part
#
# Requires a surface "BotSurf" be defined which defines source face of extrusion
# Automatically creates opposite surface "TopSurf"

import math, copy
from types import *
from abaqus import *

class PmeRegions:
    def __init__(self):
        pass

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRegionSequenceFromSet(self, p, regionType, partSetName):
        if not p.sets.has_key(partSetName): return []
        pRegions = getattr(p, regionType)
        regions = pRegions[0:0]
        for region in getattr(p.sets[partSetName], regionType):
            regions = regions + pRegions[region.index:region.index+1]
        return regions

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRegionSequenceFromSurface(self, p, surfaceName):
        if not p.surfaces.has_key(surfaceName): return []
        surface = p.surfaces[surfaceName]
        if len(surface.faces)==0:
            # Orphan mesh part
            pFaces = p.elementFaces
            faces = pFaces[0:0]
            pairs = {}
            for kface in range(len(surface.faces)):
                label = surface.elements[kface]
                if not pairs.has_key(label):
                    pairs[label] = []
                pairs[label].append(surface.sides[kface])
            for kelementFace in range(len(pFaces)):
                elementFace = pFaces[kelementFace]
                label = elementFace.label
                if label in pairs:
                    face = elementFace.face                    
                    if face in pairs[label]:
                        faces = faces + pFaces[kelementFace:kelementFace+1]
        else:
            # Geometry based part
            pFaces = p.faces
            faces = pFaces[0:0]
            for face in p.surfaces[surfaceName].faces:
                faces = faces + pFaces[face.index:face.index+1]
        return faces

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRegionIdsFromRegion(self, p, fromRegionType, fromRegionName, toRegionType):
        # fromRegionType = 'surfaces' or 'sets'
        # toRegionType = 'edges', 'vertices', 'faces', 'cells'
        fromRegions = getattr(p, fromRegionType)
        regionIds = []
        if fromRegions.has_key(fromRegionName):
            for fromRegion in getattr(fromRegions[fromRegionName], toRegionType):
                regionIds.append(fromRegion.index)
        return regionIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getOutOfPlaneEdgeIds(self, p, faceId=None):
        # Purpose: Create a map of faceId to a list of edgeIds which are attached
        #          to the face at only on vertex
        # returns {faceId:[edgeId1, edgeId2, ..]}
        #   for faceId=None - return entire map
        #   for faceId=Id   - return list of edge ids for single face id
        #   for faceId=[Id1, Id2, ...] - return list of edge ids for all faces in list
        #
        outOfPlaneEdgeIds = {}
        for face in p.faces:
            if type(faceId) == IntType:
                if face.index != faceId: continue
            elif type(faceId) == ListType:
                if face.index not in faceId: continue
            if len(face.getCells()) > 1: continue # Internal Face
            singleVertEdges = []
            faceVertIds = face.getVertices()
            faceEdgeIds = face.getEdges()
            for edge in p.edges:
                if edge.index not in (faceEdgeIds):
                    for vertId in edge.getVertices():
                        if vertId in faceVertIds:
                            singleVertEdges.append(edge.index)
                            break
            singleVertEdges.sort()
            outOfPlaneEdgeIds[face.index] = singleVertEdges
        if type(faceId) == IntType:
            return outOfPlaneEdgeIds[faceId]
        elif type(faceId) == ListType:
            edgeIds = []
            for faceId,edges in outOfPlaneEdgeIds.items():
                edgeIds = edgeIds + edges
            return edgeIds
        else:
            return outOfPlaneEdgeIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getFaceIdsOfOppositeSurface(self, p, surfaceName):
        oppositeFaceIds = []
        faceIds = []
        for face in p.surfaces[surfaceName].faces:
            faceIds.append(face.index)
        attachedCellIds = {}
        for faceId in faceIds:
            cellIds = p.faces[faceId].getCells()
            for cellId in cellIds:
                attachedCellIds[cellId] = None
        # Determine faces not attached to surface passed in
        unattachedFaceIds = []
        for cellId in attachedCellIds.keys():
            for curFaceId in p.cells[cellId].getFaces():
                if curFaceId in faceIds: continue
                isAttached = False
                for faceId in faceIds:
                    if self.areFacesConnected(p, faceId, curFaceId):
                        isAttached = True
                if isAttached == False:
                    unattachedFaceIds.append(curFaceId)
        return unattachedFaceIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getOppositeFaceId(self, p, faceId):
        # Purpose: Return faceId of face opposite to that passed in. Assumes
        #          part is extruded with opposing face with same number of
        #          vertices and edges as face passed in.
        # Assumption: Surface is on top or bottom (exterior) of extruded cell
        #
        botFaceIds = self.getRegionIdsFromRegion(p, "surfaces", "PME_BotSurf", "faces")
        topFaceIds = self.getRegionIdsFromRegion(p, "surfaces", "PME_TopSurf", "faces")
        if faceId in botFaceIds:
            surfFaceIds = botFaceIds
        else:
            surfFaceIds = topFaceIds 
            
        unconnectedFaceIds = []
        cellId = p.faces[faceId].getCells()[0]
        cellFaceIds = p.cells[cellId].getFaces()
        for cellFaceId in cellFaceIds:
            if cellFaceId in surfFaceIds: continue
            if self.areFacesConnected(p,cellFaceId, faceId) == False:
                if cellFaceId not in unconnectedFaceIds:
                    unconnectedFaceIds.append(cellFaceId)
        # Find face with centroid closest to centroid of faceId
        cen = p.faces[faceId].getCentroid()[0]
        oppositeFaceId = None
        minDistance = 1.E20
        for unconnectedFaceId in unconnectedFaceIds:
            cent =  p.faces[unconnectedFaceId].getCentroid()[0]
            dist = (cen[0]-cent[0])**2 + (cen[1]-cent[1])**2 +  (cen[2]-cent[2])**2
            if dist < minDistance:
                minDistance = dist
                oppositeFaceId = unconnectedFaceId
        return oppositeFaceId

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getConnectingEdgeId(self, p, faceId):
        # Purpose: Return the edgeId connecting faceId and the
        #          face opposite to it.
        # Assumptions: faceId is on "PME_BotSurf"
        #
        edgeIds= self.getConnectingEdgeIds(p, faceId)
        if len(edgeIds) > 0:
            return edgeIds[0]

     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getConnectingEdgeIds(self, p, faceId):
        # Purpose: Return a list of edgeIds connecting faceId and the
        #          face opposite to faceId.
        # Assumptions: faceId is on "PME_BotSurf"
        #
        connectingEdgeIds = []
        oppositeFaceId = self.getOppositeFaceId(p, faceId)
        outOfPlaneEdgeIds = self.getOutOfPlaneEdgeIds(p, faceId)
        oppositeOutOfPlaneEdgeIds = self.getOutOfPlaneEdgeIds(p, oppositeFaceId)
        for edgeId in outOfPlaneEdgeIds:
            if edgeId in oppositeOutOfPlaneEdgeIds:
                connectingEdgeIds.append(edgeId)
        return connectingEdgeIds
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getEdgeIdFromTwoVertices(self, p, vert1, vert2):
        # Purpose: return an edge id referencing to vertices
        refVerts = [vert1, vert2]
        refVerts.sort()
        for edge in p.edges:
            verts = list(edge.getVertices())
            verts.sort()
            if verts == refVerts:
                return edge.index

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getInternalVertexIdsOfSurface(self, p, surfaceName):
        return self.getVertexIdsFromEdgeIds(p, self.getInternalEdgeIdsOfSurface(p, surfaceName))

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getPerimeterVertexIdsOfSurface(self, p, surfaceName):
        return self.getVertexIdsFromEdgeIds(p, self.getPerimeterEdgeIdsOfSurface(p, surfaceName))

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getVertexIdsFromEdgeIds(self, p, edgeIds):
        vertIds = []
        for edgeId in edgeIds:
            for vertId in p.edges[edgeId].getVertices():
                if vertId not in vertIds:
                    vertIds.append(vertId)
        return vertIds
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getInternalEdgeIdsOfSurface(self, p, surfaceName):
         # Purpose: Create a list of edgeIds that are NOT on the perimeter
         #          of the surface
         # returns [edgeId1, edgeId2, ...]
         #
         surfFaceIds = []
         for face in p.surfaces[surfaceName].faces:
             surfFaceIds.append(face.index)
         internalEdgeIds = []
         for edgeId in self.getEdgeIdsOfSurface(p, surfaceName):
             faceIds = p.edges[edgeId].getFaces()
             if len(faceIds) > 1:
                 hasOutOfSurfaceFace = False
                 for faceId in faceIds:
                     if faceId not in surfFaceIds:
                         hasOutOfSurfaceFace = True
                 if hasOutOfSurfaceFace == False:
                     internalEdgeIds.append(edgeId)
         return internalEdgeIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getPerimeterEdgeIdsOfSurface(self, p, surfaceName):
         # Purpose: Create a list of edgeIds that are on the perimeter
         #          of the surface
         # returns [edgeId1, edgeId2, ...]
         #
         surfFaceIds = []
         for face in p.surfaces[surfaceName].faces:
             surfFaceIds.append(face.index)
         externalEdgeIds = []
         for edgeId in self.getEdgeIdsOfSurface(p, surfaceName):
             faceIds = p.edges[edgeId].getFaces()
             if len(faceIds) == 1:
                 externalEdgeIds.append(edgeId)
             else:
                 hasOutOfSurfaceFace = False
                 for faceId in faceIds:
                     if faceId not in surfFaceIds:
                         hasOutOfSurfaceFace = True
                 if hasOutOfSurfaceFace == True:
                     externalEdgeIds.append(edgeId)
         return externalEdgeIds
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getEdgeIdsOfSurface(self, p, surfaceName):
        # Purpose: Create a list of edgeIds that belong to a surface
        # returns [edgeId1, edgeId2, ...]
        surfaceEdgeIds = []
        for face in p.surfaces[surfaceName].faces:
            edgeIds = face.getEdges()
            for edgeId in face.getEdges():
                if edgeId not in surfaceEdgeIds:
                    surfaceEdgeIds.append(edgeId)
        return surfaceEdgeIds
    
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getConnectedFaceIds(self, p, faceId):
        # Purpose: return EXTERIOR faces which are connected to faceId
        localEdges = p.faces[faceId].getEdges()
        faceIds = []
        for face in p.faces:
            if face.index == faceId: continue
            if len(face.getCells())!=1: continue # Exterior surface check
            for edgeId in p.faces[face.index].getEdges():
                if edgeId in localEdges:
                    faceIds.append(face.index)
                    break
        return faceIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getSequenceFromIds(self, p, regionType, ids):
        # regionType = 'edges' or 'faces' or 'cells' or 'vertices'
        pRegions = getattr(p, regionType)
        sequence = None
        for id in ids:
            if sequence:
                sequence = sequence + pRegions[id:id+1]
            else:
                sequence = pRegions[id:id+1]
        return sequence
  
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def areFacesConnected(self, p, faceId1, faceId2):
        edgeIds1 = p.faces[faceId1].getEdges()
        edgeIds2 = p.faces[faceId2].getEdges()
        for edgeId in edgeIds1:
            if edgeId in edgeIds2:
                return True
        return False

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def pairOppositeEdgeLists(self, p, edgeIds1, edgeIds2):
        edgePairs = []
        cent1 = {}
        cent2 = {}
        for edgeId in edgeIds1:
            cent1[edgeId] = self.getEdgeCentroid(p, edgeId)
        for edgeId in edgeIds2:
            cent2[edgeId] = self.getEdgeCentroid(p, edgeId)
        for edgeId in edgeIds1:
            c1 = cent1[edgeId]
            minDistance=1.E20; minId = -1
            for id in edgeIds2:
                c2 = cent2[id]
                d = (c1[0]-c2[0])**2 + (c1[1]-c2[1])**2 + (c1[2]-c2[2])**2 
                if d < minDistance:
                    minDistance = d
                    minId = id
            edgePairs.append( (edgeId,minId) )
        return edgePairs

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def distanceBetweenVertices(self, p, vertId1, vertId2):
        c1 = p.vertices[vertId1].pointOn[0]
        c2 = p.vertices[vertId2].pointOn[0]
        distance = math.sqrt( (c1[0]-c2[0])**2 + (c1[1]-c2[1])**2 + (c1[2]-c2[2])**2 )
        return distance

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getEdgeCentroid(self, p, edgeId):
        vertIds = p.edges[edgeId].getVertices()
        x = 0.0; y = 0.0; z = 0.0
        for vertId in vertIds:
            coords = p.vertices[vertId].pointOn[0]
            x = x + coords[0]
            y = y + coords[1]
            z = z + coords[2]
        x = x/float(len(vertIds))
        y = y/float(len(vertIds))
        z = z/float(len(vertIds))
        return (x,y,z)

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def groupFaceIds(self, p, faceIds):
        # Group face ids based on faces which are connected to one another
        groups = []
        remainingIds = copy.deepcopy(faceIds)
        for kId in range(len(faceIds)):
            faceId = faceIds[kId]
            if faceId not in remainingIds: continue
            group = []
            group.append(faceId)
            groupEdgeIds = list(p.faces[faceId].getEdges())
            for kPass in range(kId+1,len(faceIds)):
                numFacesAdded = 0
                for jId in range(kId+1,len(faceIds)):
                    jFaceId = faceIds[jId]
                    if jFaceId not in remainingIds: continue
                    edgeIds = list(p.faces[jFaceId].getEdges())
                    overlap = self.getListOverlap(p, edgeIds,groupEdgeIds)
                    if len(overlap) > 0:
                        remainingIds.remove(jFaceId)
                        for id in overlap:
                            edgeIds.remove(id)
                            groupEdgeIds.remove(id)
                        group.append(jFaceId)
                        groupEdgeIds = groupEdgeIds + edgeIds
                        numFacesAdded = numFacesAdded + 1
                if numFacesAdded == 0: break
            groups.append(group)
        return groups

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRegionListFromPointOns(self, p, regionType, points, printWarning=True):
        return list(self.getRegionSequenceFromPointOns(p, regionType, points, printWarning))

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getRegionSequenceFromPointOns(self, p, regionType, points, printWarning=True):
        # points = ( (x,y,z), ...)
        # regionType = 'edges', 'cells', 'faces', 'vertices'
        pRegions = getattr(p, regionType)
        regions = pRegions[0:0]
        for point in points:
            region = pRegions.findAt(point, printWarning=printWarning)
            if region:
                regions = regions + pRegions[region.index:region.index+1]
        return regions

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getEdgeIdsFromFaceIds(self, p, faceIds):
        allEdgeIds = []
        for faceId in faceIds:
            edgeIds = p.faces[faceId].getEdges()
            for edgeId in edgeIds:
                if not edgeId in allEdgeIds:
                    allEdgeIds.append(edgeId)
        return allEdgeIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getFreeEdgeIdsFromFaceIds(self, p, faceIds):
        freeEdgeIds = list(p.faces[faceIds[0]].getEdges())
        for k in range(1,len(faceIds)):
            faceId = faceIds[k]
            edgeIds = p.faces[faceId].getEdges()
            for edgeId in edgeIds:
                if edgeId in freeEdgeIds:
                    freeEdgeIds.remove(edgeId)
                else:
                     freeEdgeIds.append(edgeId)
        return freeEdgeIds

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def getListOverlap(self, p, shortlist, longlist):
        overlap = []
        for item in shortlist:
            if item in longlist:
                overlap.append(item)
        return overlap
